import re
from math import isclose
from queue import Queue
from subprocess import Popen
from threading import Thread
from typing import List, Optional

from loguru import logger

from .event import SilenceEvent

SILENCE_DETECT_RE = re.compile(
    r"\[silencedetect.*\] silence_(start|end): (\d+(?:\.\d+)?)(?: \| silence_duration: (\d+\.\d+))?"
)

# https://ffmpeg.org/ffmpeg-filters.html#silencedetect
DEFAULT_SILENCE_DETECT_NOISE: str = "-60dB"
# https://ffmpeg.org/ffmpeg-utils.html#time-duration-syntax
DEFAULT_SILENCE_DETECT_DURATION: str = "2"


def parse_silence_detect(line: str) -> Optional[SilenceEvent]:
    match = SILENCE_DETECT_RE.search(line)
    if match is None:
        return None

    return SilenceEvent(
        kind=match.group(1),
        seconds=float(match.group(2)),
        duration=float(match.group(3)) if match.group(3) else None,
    )


def validate_silence_duration(
    start: SilenceEvent,
    end: Optional[SilenceEvent],
) -> Optional[bool]:
    if end is None:
        return None

    if end.duration is None or end.seconds is None or start.seconds is None:
        return False

    duration_ffmpeg = end.duration
    duration_ours = end.seconds - start.seconds
    is_valid = isclose(duration_ffmpeg, duration_ours, abs_tol=0.01)
    if not is_valid:
        logger.error(
            f"computed duration '{duration_ours}' "
            f"differs from ffmpeg duration '{duration_ffmpeg}'"
        )
    return is_valid


# pylint: disable=too-few-public-methods
class SilenceHandler:
    event_queue: Queue

    noise: str
    duration: str

    def __init__(
        self,
        event_queue: Queue,
        noise: str = DEFAULT_SILENCE_DETECT_NOISE,
        duration: str = DEFAULT_SILENCE_DETECT_DURATION,
    ):
        self.event_queue = event_queue
        self.noise = noise
        self.duration = duration

    def ffmpeg_output(self):
        return [
            *("-af", f"silencedetect=noise={self.noise}:d={self.duration}"),
            *("-f", "null", "/dev/null"),
        ]

    def parse_process_output(self, process: Popen):
        previous = None

        logger.debug("starting to parse stdout")
        for line in process.stderr:  # type: ignore
            event = parse_silence_detect(line)
            if event is None:
                continue

            if event.kind == "end" and previous is not None:
                validate_silence_duration(previous, event)

            logger.info(f"silence {event.kind}: {event.when}")
            self.event_queue.put(event)
            previous = event

    def process_handler(self, threads: List[Thread], process: Popen):
        thread = Thread(target=self.parse_process_output, args=(process,))
        thread.start()
        threads.append(thread)



import os
from pathlib import Path
from typing import Any

from boto3 import client
from botocore.exceptions import ClientError
from loguru import logger


# pylint: disable=too-few-public-methods
class S3ArchiveStorage:
    bucket: str
    _client: Any

    def __init__(self, url: str) -> None:
        self.bucket = url.split("//")[1]
        self._client = client(
            "s3",
            region_name=os.getenv("AWS_S3_REGION_NAME"),
            endpoint_url=os.getenv("AWS_S3_ENDPOINT_URL"),
        )

    def ingest_segment(self, tmp_segment: Path, segment_filepath: Path):
        logger.debug(f"uploading segment {tmp_segment} to s3://{self.bucket}")
        try:
            self._client.upload_file(
                str(tmp_segment),
                self.bucket,
                Key=str(segment_filepath),
            )
        except ClientError as exception:
            logger.error(f"could not upload {tmp_segment}: {exception}")
            return

        # Only remove if upload succeeded
        tmp_segment.unlink()
